from cfn_check import Collection, Rule


class ResourcesChecks(Collection):
    
    @Rule("Resources", "Resources is not empty")
    def validate_resources(self, value: dict):
        assert value is not None
        assert isinstance(value, dict)

    @Rule("Resources::*::Type", "Resources::*::Type is correctly definined")
    def validate_resource_type(self, value: str): 
        assert value is not None, '❌ Resource Type not defined'
        assert isinstance(value, str), '❌ Resource Type not a string'

    @Rule(
        "Resources::LambdaExecutionRole::Properties::Policies::[]",
        "Lambad Execution Role Policies are defined",
    )
    def validate_lambda_execution_role_policies(self, value: list[dict]):
        assert isinstance(value, list), '❌ Lambda execution roles is not a list'
        assert len(value) > 0, '❌ No policies specified'

    @Rule(
        "Resources::LambdaExecutionRole::Properties::Policies::[*]::(Policy*)",
        "Lambad Execution Role Policies are defined",
    )
    def validate_lambda_execution_role_policy_defined(self, value: str):
        assert value is not None, '❌ Lambda execution role policy not defined'

    @Rule(
        "Resources::AppendItemToListFunction::Properties::Code::ZipFile::[*]::[]",
        "Lambda List Function ZipFile Code is defined",
    )
    def validate_lambda_code_zipfile_is_defined(self, value: list[str]):
        assert value is not None, '❌ Lambda zipfile code is not defined'
        assert isinstance(value, list), '❌ Lambda execution zipfile code not a list'
        assert len(value) > 0,'❌ Lambda execution zipfile code empty'
    
    @Rule(
        "Resources::SecurityGroup::Properties::(SecurityGroup)::[]",
        "It checks Security Groups are correctly definined",
    )
    def validate_security_groups(self, value: list[dict]):
      assert len(value) > 0
      
      for item in value:
        protocol = item.get("IpProtocol")
        assert isinstance(protocol, str)
        assert protocol == "tcp"
    
        from_port = item.get("FromPort")
        assert isinstance(from_port, int)
        assert from_port == 80

        to_port = item.get('ToPort')
        assert isinstance(to_port, int)
        assert to_port == 80

        cidr_ip = item.get('CidrIp')
        assert isinstance(cidr_ip, str)
        assert cidr_ip == '0.0.0.0/0'

    @Rule(
        "Resources::AppendItemToListFunction::Properties::Code::ZipFile::[[]]",
        "Lambda List Function ZipFile Code is defined",
    )
    def validate_lambda_code_zipfile(self, value: list[str]):
        assert value is not None, '❌ Lambda zipfile code is not defined'
        assert isinstance(value, list), '❌ Lambda execution zipfile code not a list'
        assert len(value) > 0,'❌ Lambda execution zipfile code empty'

    @Rule(
        "Resources::SecurityGroup::Properties::(SecurityGroup)::[IpProtocol]",
        "It checks Security Groups IpProtocols are tcp",
    )
    def validate_ip_protocols(self, ip_protocol: str):
        assert ip_protocol is not None
        assert isinstance(ip_protocol, str)
        assert ip_protocol == "tcp"


import asyncio
import sys

from cocoa.cli import CLI, CLIStyle
from cocoa.ui.config.mode import TerminalMode
from cocoa.ui.components.terminal import Section, SectionConfig
from cocoa.ui.components.header import Header, HeaderConfig
from cocoa.ui.components.terminal import Terminal, EngineConfig

from .render import render
from .validate import validate

async def create_header(
    terminal_mode: TerminalMode = "full", 
):
    
    cfn_check_terminal_mode = "extended"
    if terminal_mode == "ci":
        cfn_check_terminal_mode = "compatability"

    header = Section(
        SectionConfig(height="xx-small", width="large"),
        components=[
            Header(
                "header",
                HeaderConfig(
                    header_text="cfn-check",
                    color="indian_red_3",
                    attributes=["bold"],
                    terminal_mode=cfn_check_terminal_mode,
                    font='cyberpunk'
                ),
            ),
        ],
    )

    terminal = Terminal(
        [
            header,
        ],
        config=EngineConfig(
            max_height=40,
            max_width=120
        )
    )

    return await terminal.render_once()



@CLI.root( 
    render, 
    validate,
    global_styles=CLIStyle(
        header=create_header,
        flag_description_color="white",
        error_color="indian_red_3",
        error_attributes=["italic"],
        flag_color="indian_red_3",
        text_color="slate_blue_2",
        subcommand_color="slate_blue_2",
        indentation=5,
        terminal_mode="extended",
    ),
)
async def cfn_check():
    '''
    Check and test CloudFormation
    '''



def run():
    asyncio.run(CLI.run(sys.argv[1:]))


from __future__ import annotations
import re
import sys
from collections import deque
from typing import Deque
from cfn_check.shared.types import Data, Items
from .token_type import TokenType


class Token:

    def __init__(
        self,
        selector: tuple[int, int] | int | re.Pattern | str,
        selector_type: TokenType,
        nested: list[Token] | None = None
    ):
        self.selector = selector
        self.selector_type = selector_type
        self._nested = nested

    def match(
        self,
        node: Data,
    ):
        if isinstance(node, dict) and self.selector_type not in [
            TokenType.WILDCARD,
        ]:
            return None, list(node.items())

        elif isinstance(node, list) and self.selector_type not in [
            TokenType.BOUND_RANGE,
            TokenType.INDEX,
            TokenType.PATTERN_RANGE,
            TokenType.UNBOUND_RANGE,
            TokenType.KEY_RANGE,
            TokenType.WILDCARD,
            TokenType.WILDCARD_RANGE,
            TokenType.NESTED_RANGE,
        ]:
            return None, node

        match self.selector_type:

            case TokenType.BOUND_RANGE:
                return self._match_bound_range(node)

            case TokenType.INDEX:
                return self._match_index(node)

            case TokenType.KEY:
                return self._match_key(node)
            
            case TokenType.KEY_RANGE:
                return self._match_key_range(node)
            
            case TokenType.NESTED_RANGE:
                return self._match_nested_range(node)
        
            case TokenType.PATTERN:
                return self._match_pattern(node)

            case TokenType.PATTERN_RANGE:
                return self._match_pattern_range(node)

            case TokenType.UNBOUND_RANGE:
                return self._match_unbound_range(node)

            case TokenType.WILDCARD:
                return self._match_wildcard(node)
            
            case TokenType.WILDCARD_RANGE:
                return self._match_wildcard_range(node)

            case _:
                return None, None

    def _match_bound_range(
        self,
        node: Data,
    ):
        if not isinstance(node, list) or not isinstance(self.selector, tuple):
            return None, None
        
        start, stop = self.selector

        if stop == sys.maxsize:
            stop = len(node)

        return [f'{start}-{stop}'], [node[start:stop]]
    
    def _match_index(
        self,
        node: Data,
    ):
        if (
            isinstance(node, list)
        ) and (
            isinstance(self.selector, int)
        ) and self.selector < len(node):
            return [str(self.selector)], [node[self.selector]]
        
        return None, None
    
    def _match_key(
        self,
        node: Data,
    ):
        
        if not isinstance(node, tuple) or len(node) < 2:
            return None, None

        key, value = node

        if key == self.selector:
            return [key], [value]
        
        return None, None
    
    def _match_pattern(
        self,
        node: Data,
    ):
        
        if not isinstance(node, tuple) or len(node) < 2:
            return None, None
        
        elif not isinstance(self.selector, re.Pattern):
            return None, None
        
        key, value = node

        if self.selector.match(key):
            return [key], [value]
        
        return None, None
    
    def _match_pattern_range(
        self,
        node: Data,
    ):
        if not isinstance(node, list) or not isinstance(self.selector, re.Pattern):
            return None, None
        
        matches = [
            (idx, item)
            for idx, item in enumerate(node)
            if self.selector.match(item) or (
                isinstance(item, (dict, list))
                and any([
                    self.selector.match(val)
                    for val in item
                ])
            )
        ]
        
        return (
            [str(idx) for idx, _ in matches],
            [item for _, item in matches]
        )
    
    def _match_unbound_range(
        self,
        node: Data,
    ):
        if not isinstance(node, list):
            return None, None

        return (
            ['[]'],
            [node],
        )
    
    def _match_key_range(
        self,
        node: Data,
    ):
        if not isinstance(node, list):
            return None, None
        
        matches = [
            (
                str(idx),
                value
            ) for idx, value in enumerate(node) if (
                str(value) == self.selector
            ) or (
                isinstance(value, (dict, list))
                and self.selector in value
            )
        ]

        for idx, match in enumerate(matches):
            match_idx, item = match

            if isinstance(item, dict):
                matches[idx] = (
                    match_idx,
                    item.get(self.selector),
                )

            elif isinstance(item, list):
                match_idx[idx] = (
                    match_idx,
                    matches[match_idx],
                )

        return (
            [str(idx) for idx, _ in matches],
            [item for _, item in matches]
        )
    
    def _match_nested_range(
        self,
        node: Data
    ):
        if not isinstance(node, list):
            return None, None
        
        keys: list[str] = []
        found: list[Data] = []

        for item in node:
            if isinstance(item, list):
                nested_keys, nested_found = self._match_nested(item)
                keys.extend([
                    f'[[{key}]]'
                    for key in nested_keys
                ])
                found.extend(nested_found)

        return (
            keys,
            found,
        )
    
    def _match_nested(
        self,
        node: Data,
    ): 
        found: Items = deque()
        keys: Deque[str] = deque()

        for token in self._nested:
            matched_keys, matches = token.match(node)

            if matched_keys and matches:
                keys.extend(matched_keys)
                found.extend(matches)

        return keys, found
    
    def _match_wildcard(
        self,
        node: Data
    ):
        if not self.selector == '*':
            return None, None
        
        if isinstance(node, dict):
            return (
                ['*' for _ in node],
                node.values()
            )
        
        elif isinstance(node, list):
            return (
                ['*' for _ in node],
                node,
            )
        
        return ['*'], [node]
    
    def _match_wildcard_range(
        self,
        node: Data
    ):
        if not self.selector == '*' or not (
            isinstance(node, list)
        ):
            return None, None
        
        return (
            ['[*]' for _ in node],
            node,
        )
        


import ruamel.yaml
import sys

class MultiTaggedObject:
    def __init__(self, value, tags):
        self.value = value
        self.tags = tags

def represent_multi_tagged_object(dumper, data):
    return dumper.represent_mapping('!MultiTagged', {'value': data.value, 'tags': data.tags})

def construct_multi_tagged_object(constructor, node):
    mapping = constructor.construct_mapping(node)
    return MultiTaggedObject(mapping['value'], mapping['tags'])

yaml = ruamel.yaml.YAML()
yaml.register_class(MultiTaggedObject)

# Example usage:
data = MultiTaggedObject("some_value", ["tag1", "tag2"])
yaml.dump({'item': data}, sys.stdout)


from cfn_check import Collection, Rule
from pydantic import BaseModel, StrictStr, StrictInt, Field
from typing import Literal

class RDSDBProperties(BaseModel):
    AvailabilityZone: StrictStr
    BackupRetentionPeriod: StrictInt
    DBInstanceClass: StrictStr = Field(pattern=r'^((db)\.(c6g|c6gd|c6gn|c6i|c6id|c7g|g5g|im4gn|is4gen|m6g|m6gd|r6g|r6gd|t4g|x2gd)\.(10xlarge|112xlarge|12xlarge|16xlarge|18xlarge|24xlarge|2xlarge|32xlarge|3xlarge|48xlarge|4xlarge|56xlarge|6xlarge|8xlarge|9xlarge|large|medium|metal|micro|nano|small|xlarge))')
    StorageType: Literal['sc1', 'st1', 'gp3']

class RDSDBInstance(BaseModel):
    Type: Literal["AWS::RDS::DBInstance"]

class EC2EbsDevice(BaseModel):
    VolumeType: StrictStr
    DeleteOnTermination: Literal[True]

class EC2BlockDeviceMappings(BaseModel):
    Ebs: EC2EbsDevice

class EC2VolumeProperties(BaseModel):
    VolumeType: Literal["sc1", "st1", "gp3"]
    BlockDeviceMappings: EC2BlockDeviceMappings

class EC2Volume(BaseModel):
    Type: Literal["AWS::EC2::Volume"]
    Properties: EC2VolumeProperties

class EC2InstanceProperties(BaseModel):
    InstanceType: StrictStr = Field(pattern=r'^((c6g|c6gd|c6gn|c6i|c6id|c7g|g5g|im4gn|is4gen|m6g|m6gd|r6g|r6gd|t4g|x2gd)\.(10xlarge|112xlarge|12xlarge|16xlarge|18xlarge|24xlarge|2xlarge|32xlarge|3xlarge|48xlarge|4xlarge|56xlarge|6xlarge|8xlarge|9xlarge|large|medium|metal|micro|nano|small|xlarge))')

class EC2Instance(BaseModel):
    Type: Literal["AWS::EC2::Instance"]
    Properties: EC2InstanceProperties

class LoggingGroupProperties(BaseModel):
    LogGroupClass: Literal["Infrequent Access"]
    RetentionInDays: StrictInt

class LoggingGroup(BaseModel):
    Type: Literal["AWS::Logs::LogGroup"]
    Properties: LoggingGroupProperties

class LambdaLoggingConfig(BaseModel):
    LogGroup: StrictStr

class LambdaProperties(BaseModel):
    LoggingConfig: LambdaLoggingConfig

class Lambda(BaseModel):
    Type: Literal["AWS::Serverless::Function", "AWS::Lambda::Function"]
    Properties: LambdaProperties

class Resource(BaseModel):
    Type: StrictStr

class ValidateResourceType(Collection):

    @Rule(
        "Resources::*",
        "It checks Resource::Type is correctly definined",
    )
    def validate_test(self, value: Resource):
        assert isinstance(value, Resource), "Not a valid CloudFormation Resource"

    @Rule(
        "Resources::*",
        "It validates a lambda is configured correctly",
        filters=[
            lambda data: data if data.get("Type") in ["AWS::Serverless::Function", "AWS::Lambda::Function"] else None,
        ]
    )
    def validate_lambda(self, lambda_resource: Lambda):
        assert isinstance(lambda_resource, Lambda), "Not a valid Lambda"
        
        resources = self.query("Resources")
        document = {}

        for resource in resources:
            document.update(resource)

        lambda_logging_group = document.get(lambda_resource.Properties.LoggingConfig.LogGroup)
        assert lambda_logging_group is not None, "No matching logging group found in Resources for Lambda"
        LoggingGroup(**lambda_logging_group)
        
    @Rule(
        "Resources::*",
        "It validates a logging group is configured correctly",
        filters=[
            lambda data: data if data.get("Type") == 'AWS::Logs::LogGroup' else None,
        ]
    )
    def validate_logging_group(self, logging_group: LoggingGroup):
        assert isinstance(logging_group, LoggingGroup), "Not a valid logging group"

    @Rule(
        "Resources::*",
        "It validates an EC2 instance is configured correctly",
        filters=[
            lambda data: data if data.get("Type") == 'AWS::EC2::Instance' else None,
        ]
    )
    def validate_ec2_instances(self, ec2_instance: EC2Instance):
        assert isinstance(ec2_instance, EC2Instance)

    @Rule(
        "Resources::*",
        "It validates an EC2 Volume is configured correctly",
        filters=[
            lambda data: data if data.get("Type") == 'AWS::EC2::Volume' else None,
        ]
    )
    def validate_ec2_volumes(self, ec2_volume: EC2Volume):
        assert isinstance(ec2_volume, EC2Volume), "Not a valid EC2 Volume"


import yaml
from cfn_check.rendering import Renderer
from cfn_check.cli.utils.files import open_template, Loader, create_tag
def test():

    renderer = Renderer()
    data = {}

    tags = [
        'Ref',
        'Sub',
        'Join',
        'Select',
        'Split',
        'GetAtt',
        'GetAZs',
        'ImportValue',
        'Equals',
        'If',
        'Not',
        'And',
        'Or',
        'Condition',
        'FindInMap',
    ]

    for tag in tags:
        new_tag = create_tag(tag)
        Loader.add_constructor(f'!{tag}', new_tag)

    _, template = open_template('template.yaml')

    data = renderer.render(template)


    with open('rendered.yaml', 'w') as yml:
        yaml.safe_dump(data, yml)




test()



from async_logging import LogLevelName, Logger, LoggingConfig
from cocoa.cli import CLI, YamlFile
from ruamel.yaml.comments import CommentedMap

from cfn_check.cli.utils.files import load_templates, write_to_file
from cfn_check.cli.utils.stdout import write_to_stdout
from cfn_check.rendering import Renderer
from cfn_check.logging.models import InfoLog
from .config import Config


@CLI.command(
    shortnames={
        'availability-zones': 'z'
    },
    display_help_on_error=False,
)
async def render(
    path: str,
    config: YamlFile[Config] = 'config.yml',
    output_file: str | None = None,
    attributes: list[str] | None = None,
    availability_zones: list[str] | None = None,
    import_values: list[str] | None = None,
    mappings: list[str] | None = None,
    parameters: list[str] | None = None,
    references: list[str] | None = None,
    log_level: LogLevelName = 'info',
):
    """
    Render a Cloud Formation template

    @param attributes A list of <key>=<value> k/v strings for !GetAtt calls to use
    @param availability-zones A list of <availability_zone> strings for !GetAZs calls to use
    @param config A CFN-Check yaml config file
    @param import-values A list of <filepath>=<export_value> k/v strings for !ImportValue 
    @param mappings A list of <key>=<value> k/v string specifying which Mappings to use
    @param output-file Path to output the rendered CloudFormation template to
    @param parameters A list of <key>=<value> k/v string for Parameters to use
    @param references A list of <key>=<value> k/v string for !Ref values to use
    @param log-level The log level to use
    """
    
    config_data = config.data
    if config_data is None:
        config_data = Config()
    
    logging_config = LoggingConfig()
    logging_config.update(
        log_level=log_level,
        log_output='stderr',
    )

    parsed_attributes: dict[str, str] | None = None
    if attributes:
        parsed_attributes = dict([
            attribute.split('=', maxsplit=1) for attribute in attributes if len(attribute.split('=', maxsplit=1)) > 0
        ])

        parsed_attributes.update(
            config_data.attributes or {}
        )

    parsed_import_values: dict[str, tuple[str, CommentedMap]] | None = None
    if import_values:
        parsed_import_value_pairs = dict([
            import_value.split('=', maxsplit=1) for import_value in import_values if len(import_value.split('=', maxsplit=1)) > 0
        ])

        parsed_import_value_pairs.update(
            config_data.import_values or {}
        )

        parsed_import_values = {}
        for import_file, import_key in parsed_import_value_pairs:
            parsed_import_values[import_file] = (
                import_key,
                await load_templates(import_file)
            )

    parsed_mappings: dict[str, str] | None = None
    if mappings:
        parsed_mappings = dict([
            mapping.split('=', maxsplit=1) for mapping in mappings if len(mapping.split('=', maxsplit=1)) > 0
        ])
        
        parsed_mappings.update(
            config_data.mappings or {}
        )

    parsed_parameters: dict[str, str] | None = None
    if parameters:
        parsed_parameters = dict([
            parameter.split('=', maxsplit=1) for parameter in parameters if len(parameter.split('=', maxsplit=1)) > 0
        ])

        parsed_parameters.update(
            config_data.parameters or {}
        )

    parsed_references: dict[str, str] | None = None
    if references:
        parsed_references = dict([
            reference.split('=', maxsplit=1) for reference in references if len(reference.split('=', maxsplit=1)) > 0
        ])

        parsed_references.update(
            config_data.references or {}
        )

    logger = Logger()

    templates = await load_templates(
        path,
    )

    assert len(templates) == 1 , '❌ Can only render one file'

    _, template = templates[0]
    renderer = Renderer()
    rendered = renderer.render(
        template,
        attributes=parsed_attributes,
        availability_zones=availability_zones,
        mappings=parsed_mappings,
        parameters=parsed_parameters,
        references=parsed_references,
    )

    if output_file is False:
        await write_to_file(output_file, rendered)
        await logger.log(InfoLog(message=f'✅ {path} template rendered'))

    else:
        await write_to_stdout(rendered)

    if config_data:
        await write_to_file(
            config.value,
            config_data.model_dump(),
        )


from async_logging import Entry, LogLevel


class DebugLog(Entry, kw_only=True):
    level: LogLevel = LogLevel.DEBUG

class InfoLog(Entry, kw_only=True):
    level: LogLevel = LogLevel.INFO

class ErrorLog(Entry, kw_only=True):
    error: str | bytes
    level: LogLevel = LogLevel.ERROR

class FatalLog(Entry, kw_only=True):
    error: str | bytes
    level: LogLevel = LogLevel.FATAL

class CriticalLog(Entry, kw_only=True):
    error: str | bytes
    level: LogLevel = LogLevel.CRITICAL


from typing import Any

from pydantic import ValidationError
from ruamel.yaml.comments import TaggedScalar, CommentedMap, CommentedSeq

from cfn_check.validation.validator import Validator
from cfn_check.shared.types import (
    YamlObject,
)

from .errors import assemble_validation_error
from .evaluator import Evaluator

class ValidationSet:

    def __init__(
        self,
        validators: list[Validator],
        flags: list[str] | None = None,
        attributes: dict[str, Any] | None = None,
        availability_zones: list[str] | None = None,
        import_values: dict[str, tuple[str, CommentedMap]] | None = None,
        mappings: dict[str, str] | None = None,
        parameters: dict[str, Any] | None = None,
        references: dict[str, str] | None = None,
    ):
        
        if flags is None:
            flags = []

        self._evaluator = Evaluator(flags=flags)
        self._validators = validators

        self._attributes: dict[str, str] | None = attributes
        self._availability_zones: list[str] | None = availability_zones
        self._mappings: dict[str, str] | None = mappings
        self._import_values: dict[
            str,
            CommentedMap | CommentedSeq | TaggedScalar | YamlObject,
        ] | None = import_values
        self._parameters: dict[str, str] | None = parameters
        self._references: dict[str, str] | None = references

    @property
    def count(self):
        return len(self._validators)

    def validate(
        self,
        templates: list[str],
    ):
        errors: list[Exception | ValidationError] = []

        for template in templates:
            for validator in self._validators:
                if errs := self._match_validator(
                    validator,
                    template,
                ):
                    errors.extend([
                        (
                            validator,
                            err
                        ) for err in errs
                    ])

        if validation_error := assemble_validation_error(errors):
            return validation_error 

    def _match_validator(
        self,
        validator: Validator,
        template: str,
    ):
        found = self._evaluator.match(
            template, 
            validator.query,
        )

        assert len(found) > 0, f"❌ No results matching results for query {validator.query}"

        errors: list[Exception | ValidationError] = []


        for matched in found:
            if err := validator(matched):
                errors.append(err)

        if len(errors) > 0:
            return errors
        